import numpy as np
import scipy.sparse.linalg as linalg
from math import sqrt, pi
import pylab as pl

def jacobi(phi, rho, h, tol=0.001):
    error = 10000.0  
    niter = 0
    while error > tol:
        niter += 1
        phi_old = phi.copy()
        phi[1:-1] = ( phi[2:] + phi[0:-2] + h**2 * rho[1:-1] ) / 2.0
        error = np.linalg.norm(phi - phi_old)
        if niter > 5000:
            print "Failed to converge"
            break
    return niter

def laplacian(x, h):
    """Calculates Laplacian of x with a second order
        finite difference method. A boundary condition x=0 is 
        assumed."""

    b = x / h**2
    c = -2*b
    c[1:] += b[:-1]
    c[:-1] += b[1:]
    return c

def conjugate_gradient(phi, rho, h, tol=0.001):
    r = -rho - laplacian(phi, h)
    p = r.copy()
    niter = 0
    while True:
        niter += 1
        z = laplacian(p,h)
        r2 = np.dot(r,r)
        alpha = r2 / np.dot(p,z)
        phi += alpha * p
        r -= alpha * z
        error = np.linalg.norm(r)
        if error < tol:
            break
        elif niter > 1000:
            print "Failed to converge"
            break
        beta = np.dot(r, r) / r2
        p = r + beta * p
    return niter
        

# Computational parameters
a = 0
b = 1.0
h = 0.02
# Boundary condition of phi(0)=phi(1.0) is taken care by the function
# laplacian, the boundary points are excluded below
x = np.arange(a+h, b, h)
N = len(x)

# source
rho = 2 * x**2 - 1

# Analytical solution
phi_correct = -x/3.0 + x**2 / 2.0 - x**4 / 6.0

# Numerical solutions
phi1 = np.zeros(N, float)
phi2 = np.zeros(N, float)

niter1 = jacobi(phi1, rho, h, 1e-4)
print "Jacobi converged in {0} iterations".format(niter1)
print "Error to analytic solution", np.linalg.norm(phi1 - phi_correct)

niter2 = conjugate_gradient(phi2, rho, h, 1e-4)
print "CG converged in {0} iterations".format(niter2)
print "Error to analytic solution", np.linalg.norm(phi2 - phi_correct)

def mv(x):
    return laplacian(x, h)

A = linalg.LinearOperator((N, N), mv, dtype=float)
phi3 = linalg.cg(A, -rho)
print "Scipy"
print "Error to analytic solution", np.linalg.norm(phi3[0] - phi_correct)

pl.plot(x, phi_correct, label="Analytic")
pl.plot(x, phi1, label="Jacobi", marker='x')
pl.plot(x, phi2, label="CG", marker='o', linewidth=2, markersize=8)
pl.plot(x, phi3[0], label="Scipy CG", marker='s')
pl.legend(loc='upper left')
pl.show()
